==> cfengine-1.4.0.tar.gz: cfengine-1.4.0/src/locks.c <==
/* cfengine for GNU
 
        Copyright (C) 1995
        Free Software Foundation, Inc.
 
   This file is part of GNU cfengine - written and maintained 
   by Mark Burgess, Dept of Computing and Engineering, Oslo College,
   Dept. of Theoretical physics, University of Oslo
 
   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

*/
 

/*********************************************************************/
/*                                                                   */
/* TOOLKIT : Locks and Signals                                       */
/*                                                                   */
/*********************************************************************/

/* A log file of the run times is kept for each host separately.
   This records each atomic lock and the time at which it
   completed, for use in computing the elapsed time. The file
   format is:

   %s time:operation:operand

   Each operation (independently of operand) has a "last" inode
   which keeps the time at which is last completed, for use in
   calculating IfElapsed. The idea here is that the elapsed time
   is from the time at which the last operation of this type
   FINISHED. This is different from a lock (which is used to
   allow several sub operations to coexist). Here we are
   limiting activity in general to avoid "spamming".

   Each atomic operation (including operand) has a lock. The
   removal of this lock causes the "last" file to be updated.
   This is used to actually prevent execution of an atom
   which is already being executed. If this lock has existed
   for longer than the ExpireAfter time, the process owning
   the lock is killed and the lock is re-established. The
   lock file contains the pid. 

   This is robust to hanging locks and can be thought of as
   a garbage collection mechanism for these locks.

   Last files are just inodes (empty files) so they use no disk.
   The locks (which never exceed the no of running processes)
   contain the pid.

   */

#include "cf.defs.h"
#include "cf.extern.h"

 /* The locks need these in case of signals */

char CFLOCK[bufsize];
char CFLOG[bufsize];
char CFLAST[bufsize];

/********************************************************************/
 
void HandleSignal(signum)
 
int signum;
 
{
printf("cfengine:%s: Received signal %s\n",VPREFIX,SIGNALS[signum]);

if (signum == SIGTERM || signum == SIGINT || signum == SIGHUP)
   {
   ReleaseCurrentLock(CFSTARTTIME);
   exit(0);
   }
}

/************************************************************************/
 
GetLock(operator,operand,ifelapsed,expireafter,host,now)

char *operator, *operand;
int ifelapsed, expireafter;
time_t now;

{ struct stat statbuf;
  int pid, lastcompleted = 0, elapsedtime;

if (IGNORELOCK)
   {
   return true;
   }

Debug("GetLock(%s,%s,time=%d)\n",operator,operand,now);

sprintf(CFLOG,"%s/cfengine.%s.runlog",VLOGDIR,host);
sprintf(CFLOCK,"%s/lock.%s.%s.%s.%s",VLOCKDIR,VCANONICALFILE,host,operator,operand);
sprintf(CFLAST,"%s/last.%s.%s.%s.%s",VLOCKDIR,VCANONICALFILE,host,operator,operand);

lastcompleted = GetLastLock();

elapsedtime = ((int)now-lastcompleted) / 60;

if (elapsedtime < ifelapsed)
   {
   Verbose("cfengine: Too soon since last run with %s (%d minutes)\n",operator,elapsedtime);
   return false;
   }
    
if (stat(CFLOCK,&statbuf) != -1)
   {
   if (elapsedtime >= expireafter)
      {
      Verbose("\ncfengine:%s: atom lock %s expired...(after %d minutes)\n",VPREFIX,CFLOCK,elapsedtime);
      
      pid = GetLockPid();

      if (pid == -1)
	 {
	 printf("cfengine:%s: illegal pid in corrupt lock %s - ignoring lock\n",VPREFIX,CFLOCK);
	 }
      else
	 {
	 Verbose("Trying to kill expired cfengine\n");
	 kill(pid,SIGCONT);
	 sleep(3);
	 kill(pid,SIGINT);
	 sleep(1);
	 kill(pid,SIGTERM);
	 sleep(5);
	 kill(pid,SIGKILL);
	 LockLog(now,pid,"Lock expired, process killed",operator,operand);
	 }

      unlink(CFLOCK);
      }
   else
      {
      Verbose("cfengine: couldn't obtain lock for %s (already running!)\n",CFLOCK);
      return false;
      }
   }

SetLock();
return true;
}
 
/************************************************************************/

ReleaseCurrentLock(now)

time_t now;

{
if (IGNORELOCK)
   {
   return;
   }

Debug("ReleaseCurrentLock(%s)\n",CFLOCK);

if (unlink(CFLOCK) == -1)
   {
   Verbose("cfengine:%s: unable to remove lock %s\n",VPREFIX,CFLOCK);
   /* perror("unlink"); */
   exit(1);
   }

if (creat(CFLAST,0644) == -1)
   {
   printf("cfengine:%s: unable to create %s\n",VPREFIX,CFLAST);
   perror("creat");
   exit(1);
   }

LockLog(now,getpid(),"Lock removed normally",CFLOCK,"");
}


/************************************************************************/
/* Level 2                                                              */
/************************************************************************/

GetLastLock()

{ struct stat statbuf;

Debug("GetLastLock()\n");

if (stat(CFLAST,&statbuf) == -1)
   {
   return 0;
   }
else
   {
   return (int) statbuf.st_mtime;
   }
}

/************************************************************************/

GetLockPid()

{ FILE *fp;
  int pid = -1;

if ((fp = fopen(CFLOCK,"r")) == NULL)
   {
   printf("cfengine:%s: GetLock weird error, lock disappeared!\n",VPREFIX);
   perror("fopen");
   exit(1);
   }

fscanf(fp,"%d",&pid);

fclose(fp);

Debug("GetLockPid return pid = %d\n",pid);
return pid;
}

/************************************************************************/

SetLock()

{ FILE *fp;

Debug("SetLock(%s)\n",CFLOCK);

if ((fp = fopen(CFLOCK,"w")) == NULL)
   {
   printf("cfengine:%s: GetLock: can't open new lock file %s\n",VPREFIX,CFLOCK);
   perror("fopen");
   exit(1);
   }

fprintf(fp,"%d\n",getpid());

fclose(fp);
}

/************************************************************************/

LockLog(tim,pid,str,operator,operand)

int pid;
time_t tim;
char *str, *operator, *operand;

{ FILE *fp;
  char buffer[maxvarsize];
  struct stat statbuf;

Debug("LockLog(%s)\n",str);

if ((fp = fopen(CFLOG,"a")) == NULL)
   {
   printf("cfengine:%s: GetLock: can't open log file %s\n",VPREFIX,CFLOG);
   perror("fopen");
   exit(1);
   }

sprintf(buffer,"%s",ctime(&tim));

Chop(buffer);

fprintf(fp,"%s:%s:pid=%d:%s:%s\n",buffer,str,pid,operator,operand);

fclose(fp);

if (stat(CFLOG,&statbuf) != -1)
   {
   if (statbuf.st_size > CFLOGSIZE)
      {
      Verbose("Rotating lock-runlog file\n");
      RotateFiles(CFLOG,2);
      }
   }
}

==> cfengine-1.4.1.tar.gz: cfengine-1.4.1/src/locks.c <==
/* cfengine for GNU
 
        Copyright (C) 1995
        Free Software Foundation, Inc.
 
   This file is part of GNU cfengine - written and maintained 
   by Mark Burgess, Dept of Computing and Engineering, Oslo College,
   Dept. of Theoretical physics, University of Oslo
 
   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

*/
 

/*********************************************************************/
/*                                                                   */
/* TOOLKIT : Locks and Signals                                       */
/*                                                                   */
/*********************************************************************/

/* A log file of the run times is kept for each host separately.
   This records each atomic lock and the time at which it
   completed, for use in computing the elapsed time. The file
   format is:

   %s time:operation:operand

   Each operation (independently of operand) has a "last" inode
   which keeps the time at which is last completed, for use in
   calculating IfElapsed. The idea here is that the elapsed time
   is from the time at which the last operation of this type
   FINISHED. This is different from a lock (which is used to
   allow several sub operations to coexist). Here we are
   limiting activity in general to avoid "spamming".

   Each atomic operation (including operand) has a lock. The
   removal of this lock causes the "last" file to be updated.
   This is used to actually prevent execution of an atom
   which is already being executed. If this lock has existed
   for longer than the ExpireAfter time, the process owning
   the lock is killed and the lock is re-established. The
   lock file contains the pid. 

   This is robust to hanging locks and can be thought of as
   a garbage collection mechanism for these locks.

   Last files are just inodes (empty files) so they use no disk.
   The locks (which never exceed the no of running processes)
   contain the pid.

   */

#include "cf.defs.h"
#include "cf.extern.h"

 /* The locks need these in case of signals */

char CFLOCK[bufsize];
char CFLOG[bufsize];
char CFLAST[bufsize];

time_t GetLastLock();
time_t CheckOldLock();

/********************************************************************/
 
void HandleSignal(signum)
 
int signum;
 
{
printf("cfengine:%s: Received signal %s\n",VPREFIX,SIGNALS[signum]);

if (signum == SIGTERM || signum == SIGINT || signum == SIGHUP)
   {
   ReleaseCurrentLock();
   exit(0);
   }
}

/************************************************************************/
 
GetLock(operator,operand,ifelapsed,expireafter,host,now)

char *operator, *operand, *host;
int ifelapsed, expireafter;
time_t now;

{ struct stat statbuf;
  unsigned int pid;
  time_t lastcompleted = 0, elapsedtime;

if (IGNORELOCK)
   {
   return true;
   }

Debug("GetLock(%s,%s,time=%d)\n",operator,operand,now);

sprintf(CFLOG,"%s/cfengine.%s.runlog",VLOGDIR,host);
sprintf(CFLOCK,"%s/lock.%s.%s.%s.%s",VLOCKDIR,VCANONICALFILE,host,operator,operand);
sprintf(CFLAST,"%s/last.%s.%s.%s.%s",VLOCKDIR,VCANONICALFILE,host,operator,operand);

/* Look for non-existent (old) processes */

lastcompleted = GetLastLock();
elapsedtime = (time_t)(now-lastcompleted) / 60;

if (elapsedtime < ifelapsed)
   {
   Verbose("cfengine: Too soon since last run with %s.%s (%u/%u minutes)\n",operator,operand,elapsedtime,ifelapsed);
   return false;
   }

/* Look for existing (current) processes */

lastcompleted = CheckOldLock();
elapsedtime = (time_t)(now-lastcompleted) / 60;
    
if (lastcompleted != 0)
   {
   if (elapsedtime >= expireafter)
      {
      Verbose("\ncfengine:%s: atom lock %s expired...(after %u/%u minutes)\n",VPREFIX,CFLOCK,elapsedtime,expireafter);
      
      pid = GetLockPid();

      if (pid == -1)
	 {
	 printf("cfengine:%s: illegal pid in corrupt lock %s - ignoring lock\n",VPREFIX,CFLOCK);
	 }
      else
	 {
	 Verbose("Trying to kill expired cfengine\n");
	 kill(pid,SIGCONT);
	 sleep(3);
	 kill(pid,SIGINT);
	 sleep(1);
	 kill(pid,SIGTERM);
	 sleep(5);
	 kill(pid,SIGKILL);
	 LockLog(pid,"Lock expired, process killed",operator,operand);
	 }

      unlink(CFLOCK);
      }
   else
      {
      Verbose("cfengine: couldn't obtain lock for %s (already running!)\n",CFLOCK);
      return false;
      }
   }

SetLock();
return true;
}
 
/************************************************************************/

ReleaseCurrentLock()

{
if (IGNORELOCK)
   {
   return;
   }

Debug("ReleaseCurrentLock(%s)\n",CFLOCK);

if (unlink(CFLOCK) == -1)
   {
   Verbose("cfengine:%s: unable to remove lock %s\n",VPREFIX,CFLOCK);
   /* perror("unlink"); */
   exit(1);
   }

if (creat(CFLAST,0644) == -1)
   {
   printf("cfengine:%s: unable to create %s\n",VPREFIX,CFLAST);
   perror("creat");
   exit(1);
   }

LockLog(getpid(),"Lock removed normally",CFLOCK,"");
}


/************************************************************************/
/* Level 2                                                              */
/************************************************************************/

time_t GetLastLock()

{ struct stat statbuf;

Debug("GetLastLock()\n");

if (stat(CFLAST,&statbuf) == -1)
   {
   
   /* Do this to prevent deadlock loops from surviving if IfElapsed > T_sched */
   
   if (creat(CFLAST,0644) == -1)
      {
      printf("cfengine:%s: unable to create %s\n",VPREFIX,CFLAST);
      perror("creat");
      exit(1);
      }   
   return 0;
   }
else
   {
   return statbuf.st_mtime;
   }
}

/************************************************************************/

time_t CheckOldLock()

{ struct stat statbuf;

Debug("CheckOldLock()\n");

if (stat(CFLOCK,&statbuf) == -1)
   {
   return 0;
   }
else
   {
   return statbuf.st_mtime;
   }
}

/************************************************************************/

GetLockPid()

{ FILE *fp;
  int pid = -1;

if ((fp = fopen(CFLOCK,"r")) == NULL)
   {
   printf("cfengine:%s: GetLock weird error, lock disappeared!\n",VPREFIX);
   perror("fopen");
   exit(1);
   }

fscanf(fp,"%d",&pid);

fclose(fp);

Debug("GetLockPid return pid = %d\n",pid);
return pid;
}

/************************************************************************/

SetLock()

{ FILE *fp;

Debug("SetLock(%s)\n",CFLOCK);

if ((fp = fopen(CFLOCK,"w")) == NULL)
   {
   printf("cfengine:%s: GetLock: can't open new lock file %s\n",VPREFIX,CFLOCK);
   perror("fopen");
   exit(1);
   }

fprintf(fp,"%d\n",getpid());

fclose(fp);
}

/************************************************************************/

LockLog(pid,str,operator,operand)

int pid;
char *str, *operator, *operand;

{ FILE *fp;
  char buffer[maxvarsize];
  struct stat statbuf;
  time_t tim;

Debug("LockLog(%s)\n",str);

if ((fp = fopen(CFLOG,"a")) == NULL)
   {
   printf("cfengine:%s: GetLock: can't open log file %s\n",VPREFIX,CFLOG);
   perror("fopen");
   exit(1);
   }

if ((tim = time((time_t *)NULL)) == -1)
   {
   Debug("Cfengine: couldn't read system clock\n");
   }

sprintf(buffer,"%s",ctime(&tim));

Chop(buffer);

fprintf(fp,"%s:%s:pid=%d:%s:%s\n",buffer,str,pid,operator,operand);

fclose(fp);

if (stat(CFLOG,&statbuf) != -1)
   {
   if (statbuf.st_size > CFLOGSIZE)
      {
      Verbose("Rotating lock-runlog file\n");
      RotateFiles(CFLOG,2);
      }
   }
}
